/*
 * - error handling (mkl: 2005/09/12)
 */
#include <libgsloop_config.h>

#include <assert.h>
#include <errno.h>
#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>

#include <sys/types.h>
#include <sys/socket.h>

#include <gsloop_server.h>
#include <gsloop_types.h>
#include <gsloop_client.h>
#include <gsloop_con.h>


static inline int
gsloop_server_attach_client_hook(struct gsloop		*gsloop,
				 struct gsloop_server	*server,
				 struct gsloop_client	*client)
{
	const struct gsloop_server_attr	*attr = &server->attr;

	if (!attr->client_attach)
		return 0;

	return attr->client_attach((gsloop_client_t *)client,
				   attr->client_attach_data);
}


int gsloop_server_attach_client(struct gsloop_server	*server)
{
	struct gsloop		*gsloop = server->gsloop;
 	struct gsloop_client	*client;
	int			fd;
	int			err;

	client = ptx_alloc(struct gsloop_client);
	if (!client) {
		err = -ENOMEM;
		goto exit_0;
	}
		
	fd = accept(server->fd, NULL, 0);
	if (fd < 0) {
		perror("accept");
		err = -errno;
		goto exit_1;
	}
				
	err = rn_prepare_fd_for_add(fd, getpid());
	if (err) {
		err = -err;	/* rn returns errno*/
		goto exit_1;
	}
			
	err = rn_add(&gsloop->rn, fd, gsloop_client_cb, client);
	if (err) {
		err = -err;	/* rn returns errno */
		goto exit_2;
	}

	client->fd		= fd;
	client->gsloop		= gsloop;
	client->close		= server->client_close;
	client->server		= server;
	client->event		= server->attr.client_event;
	client->event_data	= server->attr.client_event_data;

	gsloop_con_add_client(gsloop, client);
	
	dprintf(LEVEL_INFO, "(%s) new connection (fd %d) from client\n",
		__PRETTY_FUNCTION__, client->fd);

	gsloop_server_attach_client_hook(gsloop, server, client);

	/* check: max allowed clients */
	if (++server->active_clients == server->attr.max_clients) {
		rn_del(&gsloop->rn, server->fd);
		dprintf(LEVEL_DEBUG, "(%s) max number of clients (%d) reached, deactivating server\n",
			__PRETTY_FUNCTION__, server->active_clients);
	}

	return fd;

 exit_2:
	close(fd);
 exit_1:
	free(client);
 exit_0:
	return err;
}


int gsloop_server_callback(rn_pollevent_t	*event)
{
	struct gsloop_server	*server = event->client.data;
	int			err = 0;

	assert(server->fd == event->fd);

	if (event->revents & POLLIN) {
		err = gsloop_server_attach_client(server);
	} else {
		switch (event->revents) {
		case POLLOUT:
			break;
		default:
			printf("Event: %d\n", event->revents);
			break;
		}
	}

	return err;
}
